"""Test NetworkElementUpdate framework"""
from datetime import datetime, timezone
import unittest
from hashlib import sha256
from uuid import uuid4

from preseem_protobuf.network_poller import network_poller_pb2
from preseem import network_element_update


class TestNetworkElementUpdate(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_01(self):
        """Try setting and getting the fields."""
        neu = network_element_update.NetworkElementUpdate()

        ts = 1612927766
        neu.time = ts
        self.assertEqual(ts, neu.time)

        company_uuid = uuid4().bytes
        neu.company_uuid = company_uuid
        self.assertEqual(company_uuid, neu.company_uuid)

        element_uuid = uuid4().bytes
        neu.element_uuid = element_uuid
        self.assertEqual(element_uuid, neu.element_uuid)

        instance = 'MyInstance'
        neu.instance = instance
        self.assertEqual(instance, neu.instance)

        neu.active = True
        self.assertEqual(neu.active, True)

        neu.pingable = True
        self.assertEqual(neu.pingable, True)

        neu.errctx = network_element_update.ErrorContext()
        self.assertEqual(neu.ping_time, None)
        neu.ping_time = 120
        self.assertEqual(neu.ping_time, 120)
        neu.ping_time = pow(2, 31)
        self.assertEqual(neu.ping_time, 120)
        neu.ping_time = -1
        self.assertEqual(neu.ping_time, 120)
        neu.ping_time = None
        self.assertEqual(neu.ping_time, None)
        self.assertEqual(len(neu.errctx.errors), 2)
        neu.errctx = None

        neu.snmp_version = network_poller_pb2.SnmpVersion.V2C
        self.assertEqual(neu.snmp_version, network_poller_pb2.SnmpVersion.V2C)

        sysobjid = '1.3.6.1.4.1.43356.1.1.3'
        self.assertEqual(neu.sysobjid, None)
        neu.sysobjid = sysobjid
        self.assertEqual(neu.sysobjid, sysobjid)
        neu.sysobjid = None
        self.assertEqual(neu.sysobjid, None)

        neu.errctx = network_element_update.ErrorContext()
        self.assertEqual(neu.poll_time, None)
        neu.poll_time = 60
        self.assertEqual(neu.poll_time, 60)
        neu.poll_time = pow(2, 31)
        self.assertEqual(neu.poll_time, 60)
        neu.poll_time = -1
        self.assertEqual(neu.poll_time, 60)
        neu.poll_time = None
        self.assertEqual(neu.poll_time, None)
        self.assertEqual(len(neu.errctx.errors), 2)
        neu.errctx = None

        self.assertEqual(neu.is_topology_stale, None)
        neu.is_topology_stale = True
        self.assertTrue(neu.is_topology_stale)
        neu.is_topology_stale = False
        self.assertFalse(neu.is_topology_stale)
        neu.is_topology_stale = None
        self.assertEqual(neu.is_topology_stale, None)

        poller_version = '11.22.33'
        neu.poller_version = poller_version
        self.assertEqual(poller_version, neu.poller_version)

        # Check functionality of a simple repeated type
        self.assertEqual(list(neu.errors), [])
        neu.errors.append(network_poller_pb2.PollError.ERROR_SNMP_OFFLINE)
        neu.errors.append(network_poller_pb2.PollError.ERROR_SNMP_BAD_MIB)
        self.assertEqual(len(neu.errors), 2)
        self.assertEqual(neu.errors[0], network_poller_pb2.PollError.ERROR_SNMP_OFFLINE)
        self.assertEqual(neu.errors[1], network_poller_pb2.PollError.ERROR_SNMP_BAD_MIB)
        del neu.errors[0]
        self.assertEqual(len(neu.errors), 1)
        self.assertEqual(neu.errors[0], network_poller_pb2.PollError.ERROR_SNMP_BAD_MIB)

        # Check functionality of a message repeated type
        self.assertEqual(list(neu.peers), [])
        peer = neu.peers.add()
        self.assertEqual(len(neu.peers), 1)
        del neu.peers[0]
        self.assertEqual(len(neu.peers), 0)

        # Test IpMac
        self.assertEqual(list(neu.ip_macs), [])
        ip_mac = neu.ip_macs.add()
        self.assertEqual(len(neu.ip_macs), 1)

        ip_mac.ip_address = "1.2.3.4"
        self.assertEqual(ip_mac.ip_address, "1.2.3.4")

        ip_mac.mac_address = "0a:00:00:01:02:01"
        self.assertEqual(ip_mac.mac_address, "0a:00:00:01:02:01")

        ip_mac.errctx = network_element_update.ErrorContext()
        self.assertEqual(ip_mac.port, None)
        ip_mac.port = 12
        self.assertEqual(ip_mac.port, 12)
        ip_mac.port = -1
        self.assertEqual(ip_mac.port, 12)
        ip_mac.port = pow(2, 31) + 1
        self.assertEqual(ip_mac.port, 12)
        ip_mac.port = None
        self.assertEqual(len(ip_mac.errctx.errors), 2)
        ip_mac.errctx = None

        del neu.ip_macs[0]
        self.assertEqual(len(neu.ip_macs), 0)

        # Test Route
        self.assertEqual(list(neu.routes), [])
        route = neu.routes.add()
        self.assertEqual(len(neu.routes), 1)

        self.assertEqual(route.net, "")
        route.net = "0b:00:00:01:02:01"
        self.assertEqual(route.net, "0b:00:00:01:02:01")

        route.errctx = network_element_update.ErrorContext()
        self.assertEqual(route.if_index, None)
        route.if_index = 11
        self.assertEqual(route.if_index, 11)
        route.if_index = -1
        self.assertEqual(route.if_index, 11)
        route.if_index = pow(2, 31) + 1
        self.assertEqual(route.if_index, 11)
        route.if_index = None
        self.assertEqual(len(route.errctx.errors), 2)
        route.errctx = None

        next_hop = '0b:00:00:01:02:03'
        self.assertEqual(route.next_hop, None)
        route.next_hop = next_hop
        self.assertEqual(route.next_hop, next_hop)
        route.next_hop = None
        self.assertEqual(route.next_hop, None)

        mac_address = '0b:00:00:01:02:04'
        self.assertEqual(route.mac_address, None)
        route.mac_address = mac_address
        self.assertEqual(route.mac_address, mac_address)
        route.mac_address = None
        self.assertEqual(route.mac_address, None)

        del neu.routes[0]
        self.assertEqual(len(neu.routes), 0)

        # Test DhcpLease
        self.assertEqual(list(neu.dhcp_leases), [])
        dhcp_lease = neu.dhcp_leases.add()
        self.assertEqual(len(neu.dhcp_leases), 1)

        dhcp_lease.ip_address = "1.2.3.4"
        self.assertEqual(dhcp_lease.ip_address, "1.2.3.4")

        dhcp_lease.mac_address = "0a:00:00:01:02:01"
        self.assertEqual(dhcp_lease.mac_address, "0a:00:00:01:02:01")

        circuit_id = '0b:00:00:01:02:07'
        self.assertEqual(dhcp_lease.circuit_id, None)
        dhcp_lease.circuit_id = circuit_id
        self.assertEqual(dhcp_lease.circuit_id, circuit_id)
        dhcp_lease.circuit_id = None
        self.assertEqual(dhcp_lease.circuit_id, None)

        remote_id = '0b:00:00:01:02:08'
        self.assertEqual(dhcp_lease.remote_id, None)
        dhcp_lease.remote_id = remote_id
        self.assertEqual(dhcp_lease.remote_id, remote_id)
        dhcp_lease.remote_id = None
        self.assertEqual(dhcp_lease.remote_id, None)

        hostname = 'testhost1'
        self.assertEqual(dhcp_lease.hostname, None)
        dhcp_lease.hostname = hostname
        self.assertEqual(dhcp_lease.hostname, hostname)
        dhcp_lease.hostname = None
        self.assertEqual(dhcp_lease.hostname, None)

        del neu.dhcp_leases[0]
        self.assertEqual(len(neu.dhcp_leases), 0)

        # Test PppoeClient
        self.assertEqual(list(neu.pppoe_clients), [])
        pppoe_client = neu.pppoe_clients.add()
        self.assertEqual(len(neu.pppoe_clients), 1)

        pppoe_client.errctx = network_element_update.ErrorContext()

        pppoe_client.username = "user1"
        self.assertEqual(pppoe_client.username, "user1")

        mac_address = '0b:00:00:01:02:04'
        self.assertEqual(pppoe_client.mac_address, None)
        pppoe_client.mac_address = mac_address
        self.assertEqual(pppoe_client.mac_address, mac_address)
        pppoe_client.mac_address = None
        self.assertEqual(pppoe_client.mac_address, None)

        del neu.pppoe_clients[0]
        self.assertEqual(len(neu.pppoe_clients), 0)

        self.assertEqual(list(pppoe_client.network_prefixes), [])
        pppoe_client.network_prefixes.append('192.0.2.100')
        self.assertEqual(len(pppoe_client.network_prefixes), 1)
        self.assertEqual(pppoe_client.network_prefixes[0], '192.0.2.100')
        pppoe_client.network_prefixes.append('192.0.3.100/29')
        self.assertEqual(len(pppoe_client.network_prefixes), 2)
        self.assertEqual(pppoe_client.network_prefixes[1], '192.0.3.100/29')
        del pppoe_client.network_prefixes[1]
        self.assertEqual(len(pppoe_client.network_prefixes), 1)
        del pppoe_client.network_prefixes[0]
        self.assertEqual(len(pppoe_client.network_prefixes), 0)

        # Check copy behavior of sub-message assignment
        ne = network_element_update.NetworkElement()
        self.assertEqual(ne.management_ip, None)
        ne.management_ip = '192.0.2.2'
        neu.data = ne  # this is a copy, not a ref
        self.assertEqual(neu.data.management_ip, '192.0.2.2')
        ne.management_ip = None
        self.assertEqual(ne.management_ip, None)
        self.assertEqual(neu.data.management_ip, '192.0.2.2')

        self.assertEqual(ne.name, None)
        ne.name = 'my-name'
        self.assertEqual(ne.name, 'my-name')
        ne.name = None
        self.assertEqual(ne.name, None)

        self.assertEqual(ne.manufacturer, None)
        ne.manufacturer = 'my-mfg'
        self.assertEqual(ne.manufacturer, 'my-mfg')
        ne.manufacturer = None
        self.assertEqual(ne.manufacturer, None)

        self.assertEqual(ne.model, None)
        ne.model = 'my-mod'
        self.assertEqual(ne.model, 'my-mod')
        ne.model = None
        self.assertEqual(ne.model, None)

        self.assertEqual(ne.serial_number, None)
        ne.serial_number = 'my-ser'
        self.assertEqual(ne.serial_number, 'my-ser')
        ne.serial_number = None
        self.assertEqual(ne.serial_number, None)

        self.assertEqual(ne.system_mac_address, None)
        ne.system_mac_address = '0a:00:00:01:02:03'
        self.assertEqual(ne.system_mac_address, '0a:00:00:01:02:03')
        with self.assertRaises(ValueError):
            ne.system_mac_address = 'asdf'
        with self.assertRaises(ValueError):
            ne.system_mac_address = '11:22:33:44:55:66:77'

        ne.system_mac_address = None
        self.assertEqual(ne.system_mac_address, None)

        self.assertEqual(ne.sw_version, None)
        ne.sw_version = '1.2.3'
        self.assertEqual(ne.sw_version, '1.2.3')
        ne.sw_version = None
        self.assertEqual(ne.sw_version, None)

        ne.errctx = network_element_update.ErrorContext()
        self.assertEqual(ne.uptime, None)
        ne.uptime = 12345
        self.assertEqual(ne.uptime, 12345)
        ne.uptime = -1
        self.assertEqual(ne.uptime, 12345)
        ne.uptime = pow(2, 63) + 1
        self.assertEqual(ne.uptime, 12345)
        ne.uptime = None
        self.assertEqual(len(ne.errctx.errors), 2)
        ne.errctx = None

        ne.errctx = network_element_update.ErrorContext()
        self.assertEqual(ne.cpu, None)
        ne.cpu = 5.5
        self.assertEqual(round(ne.cpu, 4), 5.5)
        ne.cpu = -1
        self.assertEqual(round(ne.cpu, 4), 5.5)
        ne.cpu = 100.1
        self.assertEqual(round(ne.cpu, 4), 5.5)
        ne.cpu = None
        self.assertEqual(ne.cpu, None)
        self.assertEqual(len(ne.errctx.errors), 2)
        ne.errctx = None

        ne.errctx = network_element_update.ErrorContext()
        self.assertEqual(ne.mem, None)
        ne.mem = 5.6
        self.assertEqual(round(ne.mem, 4), 5.6)
        ne.mem = -1
        self.assertEqual(round(ne.mem, 4), 5.6)
        ne.mem = 100.1
        self.assertEqual(round(ne.mem, 4), 5.6)
        ne.mem = None
        self.assertEqual(ne.mem, None)
        self.assertEqual(len(ne.errctx.errors), 2)
        ne.errctx = None

        self.assertEqual(ne.temperature, None)
        ne.temperature = 31.02
        self.assertEqual(round(ne.temperature, 4), 31.02)
        ne.temperature = -10
        self.assertEqual(round(ne.temperature, 4), -10)
        ne.temperature = None
        self.assertEqual(ne.temperature, None)

        poller_hash = sha256().digest()
        ne.poller_hash = poller_hash
        self.assertEqual(poller_hash, ne.poller_hash)

        ne.errctx = network_element_update.ErrorContext()
        self.assertEqual(ne.gps_sats_tracked, None)
        ne.gps_sats_tracked = 12
        self.assertEqual(ne.gps_sats_tracked, 12)
        ne.gps_sats_tracked = -1
        self.assertEqual(ne.gps_sats_tracked, 12)
        ne.gps_sats_tracked = pow(2, 31) + 1
        self.assertEqual(ne.gps_sats_tracked, 12)
        ne.gps_sats_tracked = None
        self.assertEqual(len(ne.errctx.errors), 2)
        ne.errctx = None

        self.assertEqual(ne.latitude, None)
        ne.latitude = 43.4772382
        self.assertEqual(ne.latitude, 43.4772382)
        ne.latitude = None
        self.assertEqual(ne.latitude, None)

        self.assertEqual(ne.longitude, None)
        ne.longitude = -80.5516853
        self.assertEqual(ne.longitude, -80.5516853)
        ne.longitude = None
        self.assertEqual(ne.longitude, None)

        self.assertEqual(ne.elevation, None)
        ne.elevation = 45.67
        self.assertEqual(round(ne.elevation, 4), 45.67)
        ne.elevation = None
        self.assertEqual(ne.elevation, None)

        self.assertEqual(ne.height, None)
        ne.height = 78.90
        self.assertEqual(round(ne.height, 4), 78.90)
        ne.height = None
        self.assertEqual(ne.height, None)

        self.assertEqual(ne.imsi, None)
        ne.imsi = '123456789012345'
        self.assertEqual(ne.imsi, '123456789012345')
        ne.imsi = None
        self.assertEqual(ne.imsi, None)

        self.assertEqual(ne.admin_disabled, None)
        ne.admin_disabled = True
        self.assertEqual(ne.admin_disabled, True)
        ne.admin_disabled = None
        self.assertEqual(ne.admin_disabled, None)

        self.assertEqual(ne.extra, None)
        ne.extra = '{"key": "value"}'
        self.assertEqual(ne.extra, '{"key": "value"}')
        ne.extra = None
        self.assertEqual(ne.extra, None)

        self.assertEqual(ne.counter_source, None)
        ne.counter_source = "http"
        self.assertEqual(ne.counter_source, "http")
        ne.counter_source = None
        self.assertEqual(ne.counter_source, None)
        ne.counter_source = "snmp"
        self.assertEqual(ne.counter_source, "snmp")

        self.assertEqual(ne.offline_reason, None)
        ne.offline_reason = network_poller_pb2.OfflineReason.OFFLINE_REASON_POWER_OFF
        self.assertEqual(ne.offline_reason, network_poller_pb2.OfflineReason.OFFLINE_REASON_POWER_OFF)
        ne.offline_reason = None
        self.assertEqual(ne.offline_reason, None)
        ne.offline_reason = network_poller_pb2.OfflineReason.OFFLINE_REASON_OTHER
        self.assertEqual(ne.offline_reason, network_poller_pb2.OfflineReason.OFFLINE_REASON_OTHER)

        self.assertEqual(ne.offline_reason_detail, None)
        ne.offline_reason_detail = "dying gasp"
        self.assertEqual(ne.offline_reason_detail, "dying gasp")
        ne.offline_reason_detail = None
        self.assertEqual(ne.offline_reason_detail, None)
        ne.offline_reason_detail = "unknown"
        self.assertEqual(ne.offline_reason_detail, "unknown")

        ut_time = datetime(2025, 7, 14, 12, 34, 56, tzinfo=timezone.utc)
        self.assertEqual(ne.offline_time, None)
        ne.offline_time = ut_time
        self.assertEqual(ne.offline_time, ut_time)
        ne.offline_time = None
        self.assertEqual(ne.offline_time, None)
        ne.offline_time = ut_time
        self.assertEqual(ne.offline_time, ut_time)

        # Interface
        intf = ne.interfaces.add()
        intf.id = '1'
        self.assertEqual(intf.id, '1')

        self.assertEqual(intf.name, None)
        intf.name = 'interface 1'
        self.assertEqual(intf.name, 'interface 1')
        intf.name = None
        self.assertEqual(intf.name, None)

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.type, None)
        intf.type = 6
        self.assertEqual(intf.type, 6)
        intf.type = -1
        self.assertEqual(intf.type, 6)
        intf.type = pow(2, 31) + 1
        self.assertEqual(intf.type, 6)
        intf.type = None
        self.assertEqual(intf.type, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.mtu, None)
        intf.mtu = 1500
        self.assertEqual(intf.mtu, 1500)
        intf.mtu = -1
        self.assertEqual(intf.mtu, 1500)
        intf.mtu = pow(2, 31) + 1
        self.assertEqual(intf.mtu, 1500)
        intf.mtu = None
        self.assertEqual(intf.mtu, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.speed, None)
        intf.speed = 10000000000
        self.assertEqual(intf.speed, 10000000000)
        intf.speed = -1
        self.assertEqual(intf.speed, 10000000000)
        intf.speed = pow(2, 63) + 1
        self.assertEqual(intf.speed, 10000000000)
        intf.speed = None
        self.assertEqual(intf.speed, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        self.assertEqual(intf.mac_address, None)
        intf.mac_address = '0a:00:00:01:02:03'
        self.assertEqual(intf.mac_address, '0a:00:00:01:02:03')
        intf.mac_address = None
        self.assertEqual(intf.mac_address, None)

        self.assertEqual(intf.admin_up, None)
        intf.admin_up = True
        self.assertEqual(intf.admin_up, True)
        intf.admin_up = None
        self.assertEqual(intf.admin_up, None)

        self.assertEqual(intf.oper_up, None)
        intf.oper_up = False
        self.assertEqual(intf.oper_up, False)
        intf.oper_up = None
        self.assertEqual(intf.oper_up, None)

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.in_octets, None)
        intf.in_octets = pow(2, 31) + 1
        self.assertEqual(intf.in_octets, pow(2, 31) + 1)
        intf.in_octets = -1
        self.assertEqual(intf.in_octets, pow(2, 31) + 1)
        intf.in_octets = pow(2, 63) + 1
        self.assertEqual(intf.in_octets, pow(2, 31) + 1)
        intf.in_octets = None
        self.assertEqual(intf.in_octets, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.in_ucast_pkts, None)
        intf.in_ucast_pkts = pow(2, 32) + 1
        self.assertEqual(intf.in_ucast_pkts, pow(2, 32) + 1)
        intf.in_ucast_pkts = -1
        self.assertEqual(intf.in_ucast_pkts, pow(2, 32) + 1)
        intf.in_ucast_pkts = pow(2, 63) + 1
        self.assertEqual(intf.in_ucast_pkts, pow(2, 32) + 1)
        intf.in_ucast_pkts = None
        self.assertEqual(intf.in_ucast_pkts, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.in_nucast_pkts, None)
        intf.in_nucast_pkts = pow(2, 33) + 1
        self.assertEqual(intf.in_nucast_pkts, pow(2, 33) + 1)
        intf.in_nucast_pkts = -1
        self.assertEqual(intf.in_nucast_pkts, pow(2, 33) + 1)
        intf.in_nucast_pkts = pow(2, 63) + 1
        self.assertEqual(intf.in_nucast_pkts, pow(2, 33) + 1)
        intf.in_nucast_pkts = None
        self.assertEqual(intf.in_nucast_pkts, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.in_discards, None)
        intf.in_discards = pow(2, 34) + 1
        self.assertEqual(intf.in_discards, pow(2, 34) + 1)
        intf.in_discards = -1
        self.assertEqual(intf.in_discards, pow(2, 34) + 1)
        intf.in_discards = pow(2, 63) + 1
        self.assertEqual(intf.in_discards, pow(2, 34) + 1)
        intf.in_discards = None
        self.assertEqual(intf.in_discards, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.in_errors, None)
        intf.in_errors = pow(2, 35) + 1
        self.assertEqual(intf.in_errors, pow(2, 35) + 1)
        intf.in_errors = -1
        self.assertEqual(intf.in_errors, pow(2, 35) + 1)
        intf.in_errors = pow(2, 63) + 1
        self.assertEqual(intf.in_errors, pow(2, 35) + 1)
        intf.in_errors = None
        self.assertEqual(intf.in_errors, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.in_unknown_protos, None)
        intf.in_unknown_protos = pow(2, 36) + 1
        self.assertEqual(intf.in_unknown_protos, pow(2, 36) + 1)
        intf.in_unknown_protos = -1
        self.assertEqual(intf.in_unknown_protos, pow(2, 36) + 1)
        intf.in_unknown_protos = pow(2, 63) + 1
        self.assertEqual(intf.in_unknown_protos, pow(2, 36) + 1)
        intf.in_unknown_protos = None
        self.assertEqual(intf.in_unknown_protos, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.out_octets, None)
        intf.out_octets = pow(2, 37) + 1
        self.assertEqual(intf.out_octets, pow(2, 37) + 1)
        intf.out_octets = -1
        self.assertEqual(intf.out_octets, pow(2, 37) + 1)
        intf.out_octets = pow(2, 63) + 1
        self.assertEqual(intf.out_octets, pow(2, 37) + 1)
        intf.out_octets = None
        self.assertEqual(intf.out_octets, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.out_ucast_pkts, None)
        intf.out_ucast_pkts = pow(2, 38) + 1
        self.assertEqual(intf.out_ucast_pkts, pow(2, 38) + 1)
        intf.out_ucast_pkts = -1
        self.assertEqual(intf.out_ucast_pkts, pow(2, 38) + 1)
        intf.out_ucast_pkts = pow(2, 63) + 1
        self.assertEqual(intf.out_ucast_pkts, pow(2, 38) + 1)
        intf.out_ucast_pkts = None
        self.assertEqual(intf.out_octets, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.out_nucast_pkts, None)
        intf.out_nucast_pkts = pow(2, 39) + 1
        self.assertEqual(intf.out_nucast_pkts, pow(2, 39) + 1)
        intf.out_nucast_pkts = -1
        self.assertEqual(intf.out_nucast_pkts, pow(2, 39) + 1)
        intf.out_nucast_pkts = pow(2, 63) + 1
        self.assertEqual(intf.out_nucast_pkts, pow(2, 39) + 1)
        intf.out_nucast_pkts = None
        self.assertEqual(intf.out_octets, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.out_discards, None)
        intf.out_discards = pow(2, 40) + 1
        self.assertEqual(intf.out_discards, pow(2, 40) + 1)
        intf.out_discards = -1
        self.assertEqual(intf.out_discards, pow(2, 40) + 1)
        intf.out_discards = pow(2, 63) + 1
        self.assertEqual(intf.out_discards, pow(2, 40) + 1)
        intf.out_discards = None
        self.assertEqual(intf.out_discards, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        intf.errctx = network_element_update.ErrorContext()
        self.assertEqual(intf.out_errors, None)
        intf.out_errors = pow(2, 41) + 1
        self.assertEqual(intf.out_errors, pow(2, 41) + 1)
        intf.out_errors = -1
        self.assertEqual(intf.out_errors, pow(2, 41) + 1)
        intf.out_errors = pow(2, 63) + 1
        self.assertEqual(intf.out_errors, pow(2, 41) + 1)
        intf.out_errors = None
        self.assertEqual(intf.out_discards, None)
        self.assertEqual(len(intf.errctx.errors), 2)
        intf.errctx = None

        self.assertEqual(list(intf.ip_addresses), [])
        intf.ip_addresses.append('192.0.2.2')
        self.assertEqual(len(intf.ip_addresses), 1)
        self.assertEqual(intf.ip_addresses[0], '192.0.2.2')
        del intf.ip_addresses[0]
        self.assertEqual(len(intf.ip_addresses), 0)

        self.assertEqual(list(intf.bridged_mac_addresses), [])
        intf.bridged_mac_addresses.append('0a:00:00:0a:0b:0c')
        self.assertEqual(len(intf.bridged_mac_addresses), 1)
        self.assertEqual(intf.bridged_mac_addresses[0], '0a:00:00:0a:0b:0c')
        self.assertEqual([x for x in intf.bridged_mac_addresses], ['0a:00:00:0a:0b:0c'])
        del intf.bridged_mac_addresses[0]
        self.assertEqual(len(intf.bridged_mac_addresses), 0)

        self.assertEqual(intf.poller_hash, None)
        poller_hash = sha256().digest()
        intf.poller_hash = poller_hash
        self.assertEqual(poller_hash, intf.poller_hash)
        intf.poller_hash = None
        self.assertEqual(intf.poller_hash, None)

        self.assertEqual(intf.manufacturer, None)
        intf.manufacturer = 'intf-mfg'
        self.assertEqual(intf.manufacturer, 'intf-mfg')
        intf.manufacturer = None
        self.assertEqual(intf.manufacturer, None)

        self.assertEqual(intf.model, None)
        intf.model = 'intf-mod'
        self.assertEqual(intf.model, 'intf-mod')
        intf.model = None
        self.assertEqual(intf.model, None)

        # Link
        link = intf.links.add()
        link.mac_address = '0a:00:00:0d:0e:0f'
        self.assertEqual(link.mac_address, '0a:00:00:0d:0e:0f')

        self.assertEqual(link.oper_up, None)
        link.oper_up = False
        self.assertEqual(link.oper_up, False)
        link.oper_up = None
        self.assertEqual(link.oper_up, None)

        link.errctx = network_element_update.ErrorContext()
        self.assertEqual(link.in_octets, None)
        link.in_octets = pow(2, 31) + 1
        self.assertEqual(link.in_octets, pow(2, 31) + 1)
        link.in_octets = -1
        self.assertEqual(link.in_octets, pow(2, 31) + 1)
        link.in_octets = pow(2, 63) + 1
        self.assertEqual(link.in_octets, pow(2, 31) + 1)
        link.in_octets = None
        self.assertEqual(link.in_octets, None)
        self.assertEqual(len(link.errctx.errors), 2)
        link.errctx = None

        link.errctx = network_element_update.ErrorContext()
        self.assertEqual(link.in_ucast_pkts, None)
        link.in_ucast_pkts = pow(2, 32) + 1
        self.assertEqual(link.in_ucast_pkts, pow(2, 32) + 1)
        link.in_ucast_pkts = -1
        self.assertEqual(link.in_ucast_pkts, pow(2, 32) + 1)
        link.in_ucast_pkts = pow(2, 63) + 1
        self.assertEqual(link.in_ucast_pkts, pow(2, 32) + 1)
        link.in_ucast_pkts = None
        self.assertEqual(link.in_ucast_pkts, None)
        self.assertEqual(len(link.errctx.errors), 2)
        link.errctx = None

        link.errctx = network_element_update.ErrorContext()
        self.assertEqual(link.in_nucast_pkts, None)
        link.in_nucast_pkts = pow(2, 33) + 1
        self.assertEqual(link.in_nucast_pkts, pow(2, 33) + 1)
        link.in_nucast_pkts = -1
        self.assertEqual(link.in_nucast_pkts, pow(2, 33) + 1)
        link.in_nucast_pkts = pow(2, 63) + 1
        self.assertEqual(link.in_nucast_pkts, pow(2, 33) + 1)
        link.in_nucast_pkts = None
        self.assertEqual(link.in_nucast_pkts, None)
        self.assertEqual(len(link.errctx.errors), 2)
        link.errctx = None

        link.errctx = network_element_update.ErrorContext()
        self.assertEqual(link.in_discards, None)
        link.in_discards = pow(2, 34) + 1
        self.assertEqual(link.in_discards, pow(2, 34) + 1)
        link.in_discards = -1
        self.assertEqual(link.in_discards, pow(2, 34) + 1)
        link.in_discards = pow(2, 63) + 1
        self.assertEqual(link.in_discards, pow(2, 34) + 1)
        link.in_discards = None
        self.assertEqual(link.in_discards, None)
        self.assertEqual(len(link.errctx.errors), 2)
        link.errctx = None

        link.errctx = network_element_update.ErrorContext()
        self.assertEqual(link.in_errors, None)
        link.in_errors = pow(2, 35) + 1
        self.assertEqual(link.in_errors, pow(2, 35) + 1)
        link.in_errors = -1
        self.assertEqual(link.in_errors, pow(2, 35) + 1)
        link.in_errors = pow(2, 63) + 1
        self.assertEqual(link.in_errors, pow(2, 35) + 1)
        link.in_errors = None
        self.assertEqual(link.in_errors, None)
        self.assertEqual(len(link.errctx.errors), 2)
        link.errctx = None

        link.errctx = network_element_update.ErrorContext()
        self.assertEqual(link.in_unknown_protos, None)
        link.in_unknown_protos = pow(2, 36) + 1
        self.assertEqual(link.in_unknown_protos, pow(2, 36) + 1)
        link.in_unknown_protos = -1
        self.assertEqual(link.in_unknown_protos, pow(2, 36) + 1)
        link.in_unknown_protos = pow(2, 63) + 1
        self.assertEqual(link.in_unknown_protos, pow(2, 36) + 1)
        link.in_unknown_protos = None
        self.assertEqual(link.in_unknown_protos, None)
        self.assertEqual(len(link.errctx.errors), 2)
        link.errctx = None

        link.errctx = network_element_update.ErrorContext()
        self.assertEqual(link.out_octets, None)
        link.out_octets = pow(2, 37) + 1
        self.assertEqual(link.out_octets, pow(2, 37) + 1)
        link.out_octets = -1
        self.assertEqual(link.out_octets, pow(2, 37) + 1)
        link.out_octets = pow(2, 63) + 1
        self.assertEqual(link.out_octets, pow(2, 37) + 1)
        link.out_octets = None
        self.assertEqual(link.out_octets, None)
        self.assertEqual(len(link.errctx.errors), 2)
        link.errctx = None

        link.errctx = network_element_update.ErrorContext()
        self.assertEqual(link.out_ucast_pkts, None)
        link.out_ucast_pkts = pow(2, 38) + 1
        self.assertEqual(link.out_ucast_pkts, pow(2, 38) + 1)
        link.out_ucast_pkts = -1
        self.assertEqual(link.out_ucast_pkts, pow(2, 38) + 1)
        link.out_ucast_pkts = pow(2, 63) + 1
        self.assertEqual(link.out_ucast_pkts, pow(2, 38) + 1)
        link.out_ucast_pkts = None
        self.assertEqual(link.out_octets, None)
        self.assertEqual(len(link.errctx.errors), 2)
        link.errctx = None

        link.errctx = network_element_update.ErrorContext()
        self.assertEqual(link.out_nucast_pkts, None)
        link.out_nucast_pkts = pow(2, 39) + 1
        self.assertEqual(link.out_nucast_pkts, pow(2, 39) + 1)
        link.out_nucast_pkts = -1
        self.assertEqual(link.out_nucast_pkts, pow(2, 39) + 1)
        link.out_nucast_pkts = pow(2, 63) + 1
        self.assertEqual(link.out_nucast_pkts, pow(2, 39) + 1)
        link.out_nucast_pkts = None
        self.assertEqual(link.out_octets, None)
        self.assertEqual(len(link.errctx.errors), 2)
        link.errctx = None

        link.errctx = network_element_update.ErrorContext()
        self.assertEqual(link.out_discards, None)
        link.out_discards = pow(2, 40) + 1
        self.assertEqual(link.out_discards, pow(2, 40) + 1)
        link.out_discards = -1
        self.assertEqual(link.out_discards, pow(2, 40) + 1)
        link.out_discards = pow(2, 63) + 1
        self.assertEqual(link.out_discards, pow(2, 40) + 1)
        link.out_discards = None
        self.assertEqual(link.out_discards, None)
        self.assertEqual(len(link.errctx.errors), 2)
        link.errctx = None

        link.errctx = network_element_update.ErrorContext()
        self.assertEqual(link.out_errors, None)
        link.out_errors = pow(2, 41) + 1
        self.assertEqual(link.out_errors, pow(2, 41) + 1)
        link.out_errors = -1
        self.assertEqual(link.out_errors, pow(2, 41) + 1)
        link.out_errors = pow(2, 63) + 1
        self.assertEqual(link.out_errors, pow(2, 41) + 1)
        link.out_errors = None
        self.assertEqual(link.out_discards, None)
        self.assertEqual(len(link.errctx.errors), 2)
        link.errctx = None

        link.errctx = network_element_update.ErrorContext()
        self.assertEqual(link.out_retransmits, None)
        link.out_retransmits = pow(2, 42) + 1
        self.assertEqual(link.out_retransmits, pow(2, 42) + 1)
        link.out_retransmits = -1
        self.assertEqual(link.out_retransmits, pow(2, 42) + 1)
        link.out_retransmits = pow(2, 63) + 1
        self.assertEqual(link.out_retransmits, pow(2, 42) + 1)
        link.out_retransmits = None
        self.assertEqual(link.out_discards, None)
        self.assertEqual(len(link.errctx.errors), 2)
        link.errctx = None

        self.assertEqual(list(link.bridged_mac_addresses), [])
        link.bridged_mac_addresses.append('1a:00:00:0a:0b:0c')
        self.assertEqual(len(link.bridged_mac_addresses), 1)
        self.assertEqual(link.bridged_mac_addresses[0], '1a:00:00:0a:0b:0c')
        self.assertEqual([x for x in link.bridged_mac_addresses], ['1a:00:00:0a:0b:0c'])
        del link.bridged_mac_addresses[0]
        self.assertEqual(len(link.bridged_mac_addresses), 0)

        poller_hash = sha256().digest()
        link.poller_hash = poller_hash
        self.assertEqual(poller_hash, link.poller_hash)

        # Radio
        radio = intf.radios.add()
        radio.id = 'radio1'
        self.assertEqual(radio.id, 'radio1')

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.frequency, None)
        radio.frequency = 5540
        self.assertEqual(radio.frequency, 5540)
        radio.frequency = -1
        self.assertEqual(radio.frequency, 5540)
        radio.frequency = pow(2, 31) + 1
        self.assertEqual(radio.frequency, 5540)
        radio.frequency = None
        self.assertEqual(radio.frequency, None)
        self.assertEqual(len(radio.errctx.errors), 2)
        radio.errctx = None

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.chan_width, None)
        radio.chan_width = 40
        self.assertEqual(radio.chan_width, 40)
        radio.chan_width = -1
        self.assertEqual(radio.chan_width, 40)
        radio.chan_width = pow(2, 31) + 1
        self.assertEqual(radio.chan_width, 40)
        radio.chan_width = None
        self.assertEqual(radio.chan_width, None)
        self.assertEqual(len(radio.errctx.errors), 2)
        radio.errctx = None

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.tx_power, None)
        radio.tx_power = 15
        self.assertEqual(radio.tx_power, 15)
        radio.tx_power = -3
        self.assertEqual(radio.tx_power, -3)
        radio.tx_power = pow(2, 31) + 1
        self.assertEqual(radio.tx_power, -3)
        radio.tx_power = None
        self.assertEqual(radio.tx_power, None)
        self.assertEqual(len(radio.errctx.errors), 1)
        radio.errctx = None

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.dl_percent, None)
        radio.dl_percent = 75
        self.assertEqual(radio.dl_percent, 75)
        radio.dl_percent = -1
        self.assertEqual(radio.dl_percent, 75)
        radio.dl_percent = pow(2, 31) + 1
        self.assertEqual(radio.dl_percent, 75)
        radio.dl_percent = None
        self.assertEqual(radio.dl_percent, None)
        self.assertEqual(len(radio.errctx.errors), 2)
        radio.errctx = None

        radio.framing = network_poller_pb2.FramingMode.FIXED
        self.assertEqual(radio.framing, network_poller_pb2.FramingMode.FIXED)

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.rx_airtime, None)
        radio.rx_airtime = 19.2
        self.assertEqual(round(radio.rx_airtime, 4), 19.2)
        radio.rx_airtime = -1
        self.assertEqual(round(radio.rx_airtime, 4), 19.2)
        radio.rx_airtime = 100.1
        self.assertEqual(round(radio.rx_airtime, 4), 100.1)
        radio.rx_airtime = None
        self.assertEqual(radio.rx_airtime, None)
        self.assertEqual(len(radio.errctx.errors), 1)
        radio.errctx = None

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.tx_airtime, None)
        radio.tx_airtime = 6.8
        self.assertEqual(round(radio.tx_airtime, 4), 6.8)
        radio.tx_airtime = -1
        self.assertEqual(round(radio.tx_airtime, 4), 6.8)
        radio.tx_airtime = 100.1
        self.assertEqual(round(radio.tx_airtime, 4), 100.1)
        radio.tx_airtime = None
        self.assertEqual(radio.tx_airtime, None)
        self.assertEqual(len(radio.errctx.errors), 1)
        radio.errctx = None

        self.assertEqual(radio.gps_sync, None)
        radio.gps_sync = True
        self.assertEqual(radio.gps_sync, True)
        radio.gps_sync = None
        self.assertEqual(radio.gps_sync, None)

        self.assertEqual(radio.chain_balance, None)
        radio.chain_balance = 0.56
        self.assertEqual(round(radio.chain_balance, 4), 0.56)
        radio.chain_balance = None
        self.assertEqual(radio.chain_balance, None)

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.rx_link_rate, None)
        radio.rx_link_rate = 7449000000
        self.assertEqual(radio.rx_link_rate, 7449000000)
        radio.rx_link_rate = -1
        self.assertEqual(radio.rx_link_rate, 7449000000)
        radio.rx_link_rate = pow(2, 63) + 1
        self.assertEqual(radio.rx_link_rate, 7449000000)
        radio.rx_link_rate = None
        self.assertEqual(radio.rx_link_rate, None)
        self.assertEqual(len(radio.errctx.errors), 2)
        radio.errctx = None

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.tx_link_rate, None)
        radio.tx_link_rate = 17449000000
        self.assertEqual(radio.tx_link_rate, 17449000000)
        radio.tx_link_rate = -1
        self.assertEqual(radio.tx_link_rate, 17449000000)
        radio.tx_link_rate = pow(2, 63) + 1
        self.assertEqual(radio.tx_link_rate, 17449000000)
        radio.tx_link_rate = None
        self.assertEqual(radio.tx_link_rate, None)
        self.assertEqual(len(radio.errctx.errors), 2)
        radio.errctx = None

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.frame_duration, None)
        radio.frame_duration = 5000
        self.assertEqual(radio.frame_duration, 5000)
        radio.frame_duration = -1
        self.assertEqual(radio.frame_duration, 5000)
        radio.frame_duration = pow(2, 63) + 1
        self.assertEqual(radio.frame_duration, 5000)
        radio.frame_duration = None
        self.assertEqual(radio.frame_duration, None)
        self.assertEqual(len(radio.errctx.errors), 2)
        radio.errctx = None

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.rx_multiplexing_gain, None)
        radio.rx_multiplexing_gain = 1.1
        self.assertEqual(round(radio.rx_multiplexing_gain, 4), 1.1)
        radio.rx_multiplexing_gain = None
        self.assertEqual(radio.rx_multiplexing_gain, None)

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.tx_multiplexing_gain, None)
        radio.tx_multiplexing_gain = 1.1
        self.assertEqual(round(radio.tx_multiplexing_gain, 4), 1.1)
        radio.tx_multiplexing_gain = None
        self.assertEqual(radio.tx_multiplexing_gain, None)

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.rx_average_group_size, None)
        radio.rx_average_group_size = 2.1
        self.assertEqual(round(radio.rx_average_group_size, 4), 2.1)
        radio.rx_average_group_size = None
        self.assertEqual(radio.rx_average_group_size, None)

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.tx_average_group_size, None)
        radio.tx_average_group_size = 2.1
        self.assertEqual(round(radio.tx_average_group_size, 4), 2.1)
        radio.tx_average_group_size = None
        self.assertEqual(radio.tx_average_group_size, None)

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.wifi_mode, None)
        radio.wifi_mode = network_poller_pb2.WifiStandard.WIFI_STANDARD_80211B
        self.assertEqual(radio.wifi_mode,
                         network_poller_pb2.WifiStandard.WIFI_STANDARD_80211B)
        radio.wifi_mode = network_poller_pb2.WifiStandard.WIFI_STANDARD_INVALID
        self.assertEqual(len(radio.errctx.errors), 1)
        self.assertEqual(radio.wifi_mode, None)
        radio.wifi_mode = network_poller_pb2.WifiStandard.WIFI_STANDARD_80211AY
        self.assertEqual(radio.wifi_mode,
                         network_poller_pb2.WifiStandard.WIFI_STANDARD_80211AY)
        radio.wifi_mode = None
        self.assertEqual(radio.wifi_mode, None)

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.max_cpe_radios, None)
        radio.max_cpe_radios = 10
        self.assertEqual(radio.max_cpe_radios, 10)
        radio.max_cpe_radios = -1
        self.assertEqual(radio.max_cpe_radios, 10)
        radio.max_cpe_radios = pow(2, 31)
        self.assertEqual(radio.max_cpe_radios, 10)
        radio.max_cpe_radios = None
        self.assertEqual(radio.max_cpe_radios, None)
        self.assertEqual(len(radio.errctx.errors), 2)
        radio.errctx = None

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.max_distance, None)
        radio.max_distance = 8000
        self.assertEqual(radio.max_distance, 8000)
        radio.max_distance = -1
        self.assertEqual(radio.max_distance, 8000)
        radio.max_distance = pow(2, 31)
        self.assertEqual(radio.max_distance, 8000)
        radio.max_distance = None
        self.assertEqual(radio.max_distance, None)
        self.assertEqual(len(radio.errctx.errors), 2)
        radio.errctx = None

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.rx_path_loss, None)
        radio.rx_path_loss = -20.59
        self.assertEqual(round(radio.rx_path_loss, 4), -20.59)
        radio.rx_path_loss = 190.59
        self.assertEqual(round(radio.rx_path_loss, 4), -20.59)
        radio.rx_path_loss = None
        self.assertEqual(radio.rx_path_loss, None)
        self.assertEqual(len(radio.errctx.errors), 1)
        radio.errctx = None

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.rx_max_interference_noise_ratio, None)
        radio.rx_max_interference_noise_ratio = -20.59
        self.assertEqual(round(radio.rx_max_interference_noise_ratio, 4), -20.59)
        radio.rx_max_interference_noise_ratio = 190.59
        self.assertEqual(round(radio.rx_max_interference_noise_ratio, 4), -20.59)
        radio.rx_max_interference_noise_ratio = None
        self.assertEqual(radio.rx_max_interference_noise_ratio, None)
        self.assertEqual(len(radio.errctx.errors), 1)
        radio.errctx = None

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.rx_max_sensitivity_loss, None)
        radio.rx_max_sensitivity_loss = -20.59
        self.assertEqual(round(radio.rx_max_sensitivity_loss, 4), -20.59)
        radio.rx_max_sensitivity_loss = 190.59
        self.assertEqual(round(radio.rx_max_sensitivity_loss, 4), -20.59)
        radio.rx_max_sensitivity_loss = None
        self.assertEqual(radio.rx_max_sensitivity_loss, None)
        self.assertEqual(len(radio.errctx.errors), 1)
        radio.errctx = None

        radio.errctx = network_element_update.ErrorContext()
        self.assertEqual(radio.azimuth, None)
        radio.azimuth = 0
        self.assertEqual(round(radio.azimuth, 4), 0)
        radio.azimuth = 180.5
        self.assertEqual(round(radio.azimuth, 4), 180.5)
        radio.max_distance = -1
        self.assertEqual(round(radio.azimuth, 4), 180.5)
        radio.azimuth = 360
        self.assertEqual(round(radio.azimuth, 4), 360)
        radio.azimuth = 361
        self.assertEqual(round(radio.azimuth, 4), 360)
        radio.azimuth = None
        self.assertEqual(radio.azimuth, None)
        self.assertEqual(len(radio.errctx.errors), 2)
        radio.errctx = None

        self.assertEqual(radio.admin_up, None)
        radio.admin_up = True
        self.assertEqual(radio.admin_up, True)
        radio.admin_up = None
        self.assertEqual(radio.admin_up, None)

        self.assertEqual(radio.oper_up, None)
        radio.oper_up = True
        self.assertEqual(radio.oper_up, True)
        radio.oper_up = None
        self.assertEqual(radio.oper_up, None)

        # RadioStream
        stream = radio.streams.add()
        stream.id = '1'
        self.assertEqual(stream.id, '1')

        self.assertEqual(stream.polarization, None)
        stream.polarization = 'HORIZONTAL'
        self.assertEqual(stream.polarization, 'HORIZONTAL')
        stream.polarization = None
        self.assertEqual(stream.polarization, None)

        self.assertEqual(stream.noise_floor, None)
        stream.noise_floor = -79
        self.assertEqual(round(stream.noise_floor, 4), -79)
        stream.noise_floor = None
        self.assertEqual(stream.noise_floor, None)

        # WirelessLink
        wlink = stream.links.add()
        wlink.mac_address = '0a:00:00:0d:0e:0f'
        self.assertEqual(wlink.mac_address, '0a:00:00:0d:0e:0f')

        wlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(wlink.distance, None)
        wlink.distance = 600
        self.assertEqual(wlink.distance, 600)
        wlink.distance = -1
        self.assertEqual(wlink.distance, 600)
        wlink.distance = pow(2, 31) + 1
        self.assertEqual(wlink.distance, 600)
        wlink.distance = None
        self.assertEqual(wlink.distance, None)
        self.assertEqual(len(wlink.errctx.errors), 2)
        wlink.errctx = None

        wlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(wlink.connected_time, None)
        wlink.connected_time = pow(2, 31) + 100
        self.assertEqual(wlink.connected_time, pow(2, 31) + 100)
        wlink.connected_time = -1
        self.assertEqual(wlink.connected_time, pow(2, 31) + 100)
        wlink.connected_time = pow(2, 63) + 1
        self.assertEqual(wlink.connected_time, pow(2, 31) + 100)
        wlink.connected_time = None
        self.assertEqual(wlink.connected_time, None)
        self.assertEqual(len(wlink.errctx.errors), 2)
        wlink.errctx = None

        self.assertEqual(wlink.noise_floor, None)
        wlink.noise_floor = -89
        self.assertEqual(round(wlink.noise_floor, 4), -89)
        wlink.noise_floor = None
        self.assertEqual(wlink.noise_floor, None)

        self.assertEqual(wlink.rssi, None)
        wlink.rssi = -51
        self.assertEqual(round(wlink.rssi, 4), -51)
        wlink.rssi = None
        self.assertEqual(wlink.rssi, None)

        self.assertEqual(wlink.snr, None)
        wlink.snr = 44
        self.assertEqual(round(wlink.snr, 4), 44)
        wlink.snr = None
        self.assertEqual(wlink.snr, None)

        wlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(wlink.rx_link_rate, None)
        wlink.rx_link_rate = 17449000000
        self.assertEqual(wlink.rx_link_rate, 17449000000)
        wlink.rx_link_rate = -1
        self.assertEqual(wlink.rx_link_rate, 17449000000)
        wlink.rx_link_rate = pow(2, 63) + 1
        self.assertEqual(wlink.rx_link_rate, 17449000000)
        wlink.rx_link_rate = None
        self.assertEqual(wlink.rx_link_rate, None)
        self.assertEqual(len(wlink.errctx.errors), 2)
        wlink.errctx = None

        wlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(wlink.tx_link_rate, None)
        wlink.tx_link_rate = 6410000000
        self.assertEqual(wlink.tx_link_rate, 6410000000)
        wlink.tx_link_rate = -1
        self.assertEqual(wlink.tx_link_rate, 6410000000)
        wlink.tx_link_rate = pow(2, 63) + 1
        self.assertEqual(wlink.tx_link_rate, 6410000000)
        wlink.tx_link_rate = None
        self.assertEqual(wlink.tx_link_rate, None)
        self.assertEqual(len(wlink.errctx.errors), 2)
        wlink.errctx = None

        wlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(wlink.rx_mcs, None)
        wlink.rx_mcs = 4
        self.assertEqual(wlink.rx_mcs, 4)
        wlink.rx_mcs = -1
        self.assertEqual(wlink.rx_mcs, 4)
        wlink.rx_mcs = pow(2, 31) + 1
        self.assertEqual(wlink.rx_mcs, 4)
        wlink.rx_mcs = None
        self.assertEqual(wlink.rx_mcs, None)
        self.assertEqual(len(wlink.errctx.errors), 2)
        wlink.errctx = None

        wlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(wlink.tx_mcs, None)
        wlink.tx_mcs = 4
        self.assertEqual(wlink.tx_mcs, 4)
        wlink.tx_mcs = -1
        self.assertEqual(wlink.tx_mcs, 4)
        wlink.tx_mcs = pow(2, 31) + 1
        self.assertEqual(wlink.tx_mcs, 4)
        wlink.tx_mcs = None
        self.assertEqual(wlink.tx_mcs, None)
        self.assertEqual(len(wlink.errctx.errors), 2)
        wlink.errctx = None

        self.assertEqual(wlink.rx_modulation_rate, None)
        wlink.rx_modulation_rate = 10.8
        self.assertEqual(round(wlink.rx_modulation_rate, 4), 10.8)
        wlink.rx_modulation_rate = None
        self.assertEqual(wlink.rx_modulation_rate, None)

        self.assertEqual(wlink.tx_modulation_rate, None)
        wlink.tx_modulation_rate = 7.7
        self.assertEqual(round(wlink.tx_modulation_rate, 4), 7.7)
        wlink.tx_modulation_rate = None
        self.assertEqual(wlink.tx_modulation_rate, None)

        wlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(wlink.rx_airtime, None)
        wlink.rx_airtime = 60.4
        self.assertEqual(round(wlink.rx_airtime, 4), 60.4)
        wlink.rx_airtime = -1
        self.assertEqual(round(wlink.rx_airtime, 4), 60.4)
        wlink.rx_airtime = 100.1
        self.assertEqual(round(wlink.rx_airtime, 4), 100.1)
        wlink.rx_airtime = None
        self.assertEqual(wlink.rx_airtime, None)
        self.assertEqual(len(wlink.errctx.errors), 1)
        wlink.errctx = None

        wlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(wlink.tx_airtime, None)
        wlink.tx_airtime = 4.5
        self.assertEqual(round(wlink.tx_airtime, 4), 4.5)
        wlink.tx_airtime = -1
        self.assertEqual(round(wlink.tx_airtime, 4), 4.5)
        wlink.tx_airtime = 100.1
        self.assertEqual(round(wlink.tx_airtime, 4), 100.1)
        wlink.tx_airtime = None
        self.assertEqual(wlink.tx_airtime, None)
        self.assertEqual(len(wlink.errctx.errors), 1)
        wlink.errctx = None

        wlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(wlink.amc, None)
        wlink.amc = 4.5
        self.assertEqual(round(wlink.amc, 4), 4.5)
        wlink.amc = -1
        self.assertEqual(round(wlink.amc, 4), 4.5)
        wlink.amc = 100.1
        self.assertEqual(round(wlink.amc, 4), 4.5)
        wlink.amc = None
        self.assertEqual(wlink.amc, None)
        self.assertEqual(len(wlink.errctx.errors), 2)
        wlink.errctx = None

        wlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(wlink.amq, None)
        wlink.amq = 4.5
        self.assertEqual(round(wlink.amq, 4), 4.5)
        wlink.amq = -1
        self.assertEqual(round(wlink.amq, 4), 4.5)
        wlink.amq = 100.1
        self.assertEqual(round(wlink.amq, 4), 4.5)
        wlink.amq = None
        self.assertEqual(wlink.amq, None)
        self.assertEqual(len(wlink.errctx.errors), 2)
        wlink.errctx = None

        wlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(wlink.ccq, None)
        wlink.ccq = 4.5
        self.assertEqual(round(wlink.ccq, 4), 4.5)
        wlink.ccq = -1
        self.assertEqual(round(wlink.ccq, 4), 4.5)
        wlink.ccq = 100.1
        self.assertEqual(round(wlink.ccq, 4), 4.5)
        wlink.ccq = None
        self.assertEqual(wlink.ccq, None)
        self.assertEqual(len(wlink.errctx.errors), 2)
        wlink.errctx = None

        poller_hash = sha256().digest()
        wlink.poller_hash = poller_hash
        self.assertEqual(poller_hash, wlink.poller_hash)

        # RfAggregationGroup
        rf_aggr_gr = intf.rf_aggregation_groups.add()
        rf_aggr_gr.name = 'RadioGr1'
        self.assertEqual(rf_aggr_gr.name, 'RadioGr1')

        rf_aggr_gr.errctx = network_element_update.ErrorContext()
        self.assertEqual(rf_aggr_gr.airtime, None)
        rf_aggr_gr.airtime = 19.2
        self.assertEqual(round(rf_aggr_gr.airtime, 4), 19.2)
        rf_aggr_gr.airtime = -1
        self.assertEqual(round(rf_aggr_gr.airtime, 4), 19.2)
        rf_aggr_gr.airtime = 100.1
        self.assertEqual(round(rf_aggr_gr.airtime, 4), 100.1)
        rf_aggr_gr.airtime = None
        self.assertEqual(rf_aggr_gr.airtime, None)
        self.assertEqual(len(rf_aggr_gr.errctx.errors), 1)
        rf_aggr_gr.errctx = None

        rf_aggr_gr.errctx = network_element_update.ErrorContext()
        self.assertEqual(rf_aggr_gr.tx_packet_error_rate, None)
        rf_aggr_gr.tx_packet_error_rate = 19.2
        self.assertEqual(round(rf_aggr_gr.tx_packet_error_rate, 4), 19.2)
        rf_aggr_gr.tx_packet_error_rate = -1
        self.assertEqual(round(rf_aggr_gr.tx_packet_error_rate, 4), 19.2)
        rf_aggr_gr.tx_packet_error_rate = 100.1
        self.assertEqual(round(rf_aggr_gr.tx_packet_error_rate, 4), 100.1)
        rf_aggr_gr.tx_packet_error_rate = None
        self.assertEqual(rf_aggr_gr.tx_packet_error_rate, None)
        self.assertEqual(len(rf_aggr_gr.errctx.errors), 1)
        rf_aggr_gr.errctx = None

        rf_aggr_gr.errctx = network_element_update.ErrorContext()
        self.assertEqual(rf_aggr_gr.rx_packet_error_rate, None)
        rf_aggr_gr.rx_packet_error_rate = 19.2
        self.assertEqual(round(rf_aggr_gr.rx_packet_error_rate, 4), 19.2)
        rf_aggr_gr.rx_packet_error_rate = -1
        self.assertEqual(round(rf_aggr_gr.rx_packet_error_rate, 4), 19.2)
        rf_aggr_gr.rx_packet_error_rate = 100.1
        self.assertEqual(round(rf_aggr_gr.rx_packet_error_rate, 4), 100.1)
        rf_aggr_gr.rx_packet_error_rate = None
        self.assertEqual(rf_aggr_gr.rx_packet_error_rate, None)
        self.assertEqual(len(rf_aggr_gr.errctx.errors), 1)
        rf_aggr_gr.errctx = None

        rf_aggr_gr.errctx = network_element_update.ErrorContext()
        self.assertEqual(rf_aggr_gr.tx_interference_signal_ratio, None)
        rf_aggr_gr.tx_interference_signal_ratio = -20.59
        self.assertEqual(round(rf_aggr_gr.tx_interference_signal_ratio, 4), -20.59)
        rf_aggr_gr.tx_interference_signal_ratio = 190.59
        self.assertEqual(round(rf_aggr_gr.tx_interference_signal_ratio, 4), 190.59)
        rf_aggr_gr.tx_interference_signal_ratio = None
        self.assertEqual(rf_aggr_gr.tx_interference_signal_ratio, None)
        self.assertEqual(len(rf_aggr_gr.errctx.errors), 0)
        rf_aggr_gr.errctx = None

        rf_aggr_gr.errctx = network_element_update.ErrorContext()
        self.assertEqual(rf_aggr_gr.rx_interference_signal_ratio, None)
        rf_aggr_gr.rx_interference_signal_ratio = -20.59
        self.assertEqual(round(rf_aggr_gr.rx_interference_signal_ratio, 4), -20.59)
        rf_aggr_gr.rx_interference_signal_ratio = 190.59
        self.assertEqual(round(rf_aggr_gr.rx_interference_signal_ratio, 4), 190.59)
        rf_aggr_gr.rx_interference_signal_ratio = None
        self.assertEqual(rf_aggr_gr.rx_interference_signal_ratio, None)
        self.assertEqual(len(rf_aggr_gr.errctx.errors), 0)
        rf_aggr_gr.errctx = None

        # Sector
        sector = intf.sectors.add()

        sector.ssid = 'MySSID'
        self.assertEqual(sector.ssid, 'MySSID')

        sector.mode = network_poller_pb2.SectorMode.AP
        self.assertEqual(sector.mode, network_poller_pb2.SectorMode.AP)

        # Module
        module = ne.modules.add()
        module.id = 'module1'
        self.assertEqual(module.id, 'module1')

        poller_hash = sha256().digest()
        module.poller_hash = poller_hash
        self.assertEqual(poller_hash, module.poller_hash)

        self.assertEqual(module.name, None)
        module.name = 'module-name'
        self.assertEqual(module.name, 'module-name')
        module.name = None
        self.assertEqual(module.name, None)

        self.assertEqual(module.manufacturer, None)
        module.manufacturer = 'module-mfg'
        self.assertEqual(module.manufacturer, 'module-mfg')
        module.manufacturer = None
        self.assertEqual(module.manufacturer, None)

        self.assertEqual(module.model, None)
        module.model = 'module-mod'
        self.assertEqual(module.model, 'module-mod')
        module.model = None
        self.assertEqual(module.model, None)

        self.assertEqual(module.serial_number, None)
        module.serial_number = 'module-ser'
        self.assertEqual(module.serial_number, 'module-ser')
        module.serial_number = None
        self.assertEqual(module.serial_number, None)

        self.assertEqual(module.sw_version, None)
        module.sw_version = '1.2.3'
        self.assertEqual(module.sw_version, '1.2.3')
        module.sw_version = None
        self.assertEqual(module.sw_version, None)

        module.errctx = network_element_update.ErrorContext()
        self.assertEqual(module.uptime, None)
        module.uptime = 12345
        self.assertEqual(module.uptime, 12345)
        module.uptime = -1
        self.assertEqual(module.uptime, 12345)
        module.uptime = pow(2, 63) + 1
        self.assertEqual(module.uptime, 12345)
        module.uptime = None
        self.assertEqual(len(module.errctx.errors), 2)
        module.errctx = None

        module.errctx = network_element_update.ErrorContext()
        self.assertEqual(module.cpu, None)
        module.cpu = 5.5
        self.assertEqual(round(module.cpu, 4), 5.5)
        module.cpu = -1
        self.assertEqual(round(module.cpu, 4), 5.5)
        module.cpu = 100.1
        self.assertEqual(round(module.cpu, 4), 5.5)
        module.cpu = None
        self.assertEqual(module.cpu, None)
        self.assertEqual(len(module.errctx.errors), 2)
        module.errctx = None

        module.errctx = network_element_update.ErrorContext()
        self.assertEqual(module.mem, None)
        module.mem = 5.6
        self.assertEqual(round(module.mem, 4), 5.6)
        module.mem = -1
        self.assertEqual(round(module.mem, 4), 5.6)
        module.mem = 100.1
        self.assertEqual(round(module.mem, 4), 5.6)
        module.mem = None
        self.assertEqual(module.mem, None)
        self.assertEqual(len(module.errctx.errors), 2)
        module.errctx = None

        self.assertEqual(module.temperature, None)
        module.temperature = 31.02
        self.assertEqual(round(module.temperature, 4), 31.02)
        module.temperature = -10
        self.assertEqual(round(module.temperature, 4), -10)
        module.temperature = None
        self.assertEqual(module.temperature, None)

        intf2 = module.interfaces.add()

        # Pon
        pon = intf2.pons.add()
        pon.id = 'pon1'
        self.assertEqual(pon.id, 'pon1')

        poller_hash = sha256().digest()
        pon.poller_hash = poller_hash
        self.assertEqual(poller_hash, pon.poller_hash)

        pon.pon_technology = network_poller_pb2.PonTechnology.APON
        self.assertEqual(pon.pon_technology, network_poller_pb2.PonTechnology.APON)
        pon.pon_technology = None
        self.assertEqual(pon.pon_technology, None)

        pon.errctx = network_element_update.ErrorContext()
        self.assertEqual(pon.split_ratio, None)
        pon.split_ratio = 5
        self.assertEqual(pon.split_ratio, 5)
        pon.split_ratio = -1
        self.assertEqual(pon.split_ratio, 5)
        pon.split_ratio = pow(2, 31) + 1
        self.assertEqual(pon.split_ratio, 5)
        pon.split_ratio = None
        self.assertEqual(pon.split_ratio, None)
        self.assertEqual(len(pon.errctx.errors), 2)
        pon.errctx = None

        pon.errctx = network_element_update.ErrorContext()
        self.assertEqual(pon.max_distance, None)
        pon.max_distance = 600
        self.assertEqual(pon.max_distance, 600)
        pon.max_distance = -1
        self.assertEqual(pon.max_distance, 600)
        pon.max_distance = pow(2, 31) + 1
        self.assertEqual(pon.max_distance, 600)
        pon.max_distance = None
        self.assertEqual(pon.max_distance, None)
        self.assertEqual(len(pon.errctx.errors), 2)
        pon.errctx = None

        self.assertEqual(pon.minimum_loss, None)
        pon.minimum_loss = 5.02
        self.assertEqual(round(pon.minimum_loss, 4), 5.02)
        pon.minimum_loss = None
        self.assertEqual(pon.minimum_loss, None)

        self.assertEqual(pon.maximum_loss, None)
        pon.maximum_loss = 6.02
        self.assertEqual(round(pon.maximum_loss, 4), 6.02)
        pon.maximum_loss = None
        self.assertEqual(pon.maximum_loss, None)

        pon.errctx = network_element_update.ErrorContext()
        self.assertEqual(pon.rx_link_rate, None)
        pon.rx_link_rate = 7449000000
        self.assertEqual(pon.rx_link_rate, 7449000000)
        pon.rx_link_rate = -1
        self.assertEqual(pon.rx_link_rate, 7449000000)
        pon.rx_link_rate = pow(2, 63) + 1
        self.assertEqual(pon.rx_link_rate, 7449000000)
        pon.rx_link_rate = None
        self.assertEqual(pon.rx_link_rate, None)
        self.assertEqual(len(pon.errctx.errors), 2)
        pon.errctx = None

        pon.errctx = network_element_update.ErrorContext()
        self.assertEqual(pon.tx_link_rate, None)
        pon.tx_link_rate = 17449000000
        self.assertEqual(pon.tx_link_rate, 17449000000)
        pon.tx_link_rate = -1
        self.assertEqual(pon.tx_link_rate, 17449000000)
        pon.tx_link_rate = pow(2, 63) + 1
        self.assertEqual(pon.tx_link_rate, 17449000000)
        pon.tx_link_rate = None
        self.assertEqual(pon.tx_link_rate, None)
        self.assertEqual(len(pon.errctx.errors), 2)
        pon.errctx = None

        self.assertEqual(pon.name, None)
        pon.name = 'pon-name'
        self.assertEqual(pon.name, 'pon-name')
        pon.name = None
        self.assertEqual(pon.name, None)

        self.assertEqual(pon.description, None)
        pon.description = 'pon-description'
        self.assertEqual(pon.description, 'pon-description')
        pon.description = None
        self.assertEqual(pon.description, None)

        # PonLink
        ponlink = pon.links.add()
        poller_hash = sha256().digest()
        ponlink.poller_hash = poller_hash
        self.assertEqual(poller_hash, ponlink.poller_hash)

        ponlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(ponlink.rx_power, None)
        ponlink.rx_power = -20.59
        self.assertEqual(round(ponlink.rx_power, 4), -20.59)
        ponlink.rx_power = 190.59
        self.assertEqual(round(ponlink.rx_power, 4), -20.59)
        ponlink.rx_power = None
        self.assertEqual(ponlink.rx_power, None)
        self.assertEqual(len(ponlink.errctx.errors), 1)
        ponlink.errctx = None

        ponlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(ponlink.tx_power, None)
        ponlink.tx_power = -20.59
        self.assertEqual(round(ponlink.tx_power, 4), -20.59)
        ponlink.tx_power = 190.59
        self.assertEqual(round(ponlink.tx_power, 4), -20.59)
        ponlink.tx_power = None
        self.assertEqual(ponlink.tx_power, None)
        self.assertEqual(len(ponlink.errctx.errors), 1)
        ponlink.errctx = None

        ponlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(ponlink.fiber_length, None)
        ponlink.fiber_length = 200
        self.assertEqual(ponlink.fiber_length, 200)
        ponlink.fiber_length = -1
        self.assertEqual(ponlink.fiber_length, 200)
        ponlink.fiber_length = pow(2, 31) + 1
        self.assertEqual(ponlink.fiber_length, 200)
        ponlink.fiber_length = None
        self.assertEqual(ponlink.fiber_length, None)
        self.assertEqual(len(ponlink.errctx.errors), 2)
        ponlink.errctx = None

        ponlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(ponlink.connected_time, None)
        ponlink.connected_time = 123456
        self.assertEqual(ponlink.connected_time, 123456)
        ponlink.connected_time = -1
        self.assertEqual(ponlink.connected_time, 123456)
        ponlink.connected_time = pow(2, 63) + 1
        self.assertEqual(ponlink.connected_time, 123456)
        ponlink.connected_time = None
        self.assertEqual(ponlink.connected_time, None)
        self.assertEqual(len(ponlink.errctx.errors), 2)
        ponlink.errctx = None

        ponlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(ponlink.rx_link_rate, None)
        ponlink.rx_link_rate = 7449000000
        self.assertEqual(ponlink.rx_link_rate, 7449000000)
        ponlink.rx_link_rate = -1
        self.assertEqual(ponlink.rx_link_rate, 7449000000)
        ponlink.rx_link_rate = pow(2, 63) + 1
        self.assertEqual(ponlink.rx_link_rate, 7449000000)
        ponlink.rx_link_rate = None
        self.assertEqual(ponlink.rx_link_rate, None)
        self.assertEqual(len(ponlink.errctx.errors), 2)
        ponlink.errctx = None

        ponlink.errctx = network_element_update.ErrorContext()
        self.assertEqual(ponlink.tx_link_rate, None)
        ponlink.tx_link_rate = 17449000000
        self.assertEqual(ponlink.tx_link_rate, 17449000000)
        ponlink.tx_link_rate = -1
        self.assertEqual(ponlink.tx_link_rate, 17449000000)
        ponlink.tx_link_rate = pow(2, 63) + 1
        self.assertEqual(ponlink.tx_link_rate, 17449000000)
        ponlink.tx_link_rate = None
        self.assertEqual(ponlink.tx_link_rate, None)
        self.assertEqual(len(ponlink.errctx.errors), 2)
        ponlink.errctx = None

    def test_add_bridged_mac(self):
        """PPA-1321 test MAC filtering when adding bridged macs via a helper method."""
        neu = network_element_update.NetworkElementUpdate()
        intf = neu.data.interfaces.add()
        link = intf.links.add()

        intf.add_bridged_mac('00:01:02:03:04:05')
        intf.add_bridged_mac('33:33:ff:00:00:00')  # should be filtered
        intf.add_bridged_mac('00:00:00:00:00:00')  # should be filtered
        intf.add_bridged_mac('00:01:02:03:04:06')
        link.add_bridged_mac('00:01:02:03:04:05')
        link.add_bridged_mac('33:33:ff:00:00:00')  # should be filtered
        link.add_bridged_mac('00:00:00:00:00:00')  # should be filtered
        link.add_bridged_mac('00:01:02:03:04:06')

        self.assertEqual(len(intf.bridged_mac_addresses), 2)
        self.assertEqual(intf.bridged_mac_addresses[0], '00:01:02:03:04:05')
        self.assertEqual(intf.bridged_mac_addresses[1], '00:01:02:03:04:06')

        self.assertEqual(len(link.bridged_mac_addresses), 2)
        self.assertEqual(link.bridged_mac_addresses[0], '00:01:02:03:04:05')
        self.assertEqual(link.bridged_mac_addresses[1], '00:01:02:03:04:06')



import yaml
import os
import logging
from collections import defaultdict


class DuplicateKeyError(Exception):
    """
    Create custom exception for duplicate keys in YAML file.
    """
    pass


class DuplicateKeySafeLoader(yaml.SafeLoader):
    """
    Extend SafeLoader from YAML loader to detect duplicate keys.
    """

    def construct_mapping(self, node, deep=False):
        """
        Override construct_mapping to detect duplicate keys in mappings.
        """
        mapping = {}
        keys = defaultdict(str)
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            if key in keys:
                line_error = f"Duplicate key '{key}' found at line {key_node.start_mark.line + 1}, "
                col_error = f"column {key_node.start_mark.column + 1}."
                logging.error(f"Ambiguous YAML detected: {line_error} {col_error}")
            keys[
                key] += "holder"  # Not value of the key because it can be a reference to a child node
            value = self.construct_object(value_node, deep=deep)
            mapping[key] = value
        return mapping


def load_config(path):
    """
    Load a YAML file and verify for ambiguity in the fields.
    """
    try:
        with open(path) as f:
            cfg = yaml.load(f.read(), Loader=DuplicateKeySafeLoader)
            return cfg
    except yaml.YAMLError as e:
        logging.error(f"Invalid YAML format: {e}")
        return None
    except FileNotFoundError:
        raise RuntimeError(f"config file not found: {path}")



"""SSH helpers"""
import asyncio
import logging

# STM-9317 suppress warnings until STM-5490 is fixed.
import warnings
from cryptography.utils import CryptographyDeprecationWarning

warnings.filterwarnings("ignore", category=CryptographyDeprecationWarning)

import asyncssh

asyncssh.set_log_level(logging.ERROR)


class NoBridgeError(Exception):
    pass


class SshClient():
    """Helper class for a SSH client.  User is responsible for retry/errors."""

    def __init__(self,
                 host,
                 port=None,
                 username=None,
                 password=None,
                 tunnel=None,
                 init_wait=0,
                 verbose=0,
                 record=False):
        self.host = host
        self.port = port or 22
        self.username = username
        self.password = password
        self.init_wait = init_wait
        self._tun = tunnel
        self.record = [] if record else None

        self.verbose = verbose
        self._conn = None
        self.stdin = self.stdout = self.stderr = None
        self.lock = None
        self.sem = asyncio.Semaphore(5)

    async def _open(self):
        if self._conn is None:
            if not self.lock:
                self.lock = asyncio.Lock()
            async with self.lock:
                if self._conn is None:
                    self._conn = await asyncio.wait_for(
                        asyncssh.connect(self.host,
                                         port=self.port,
                                         known_hosts=None,
                                         encryption_algs='+3des-cbc',
                                         kex_algs='+diffie-hellman-group1-sha1',
                                         server_host_key_algs=[
                                             'ssh-ed25519', 'rsa-sha2-256',
                                             'rsa-sha2-512', 'ssh-rsa', 'ssh-dss'
                                         ],
                                         tunnel=self._tun,
                                         username=self.username,
                                         password=self.password,
                                         preferred_auth=['password']),  # PPA-1420
                        timeout=5.0)

    async def close(self):
        """Close the connection and free any resources."""
        if self._conn:
            self._conn.close()
            self._conn = self.stdin = self.stdout = self.stderr = None

    async def tunnel(self, host, port=None, username=None, password=None):
        """Create a new client tunnelled through this one."""
        if self._conn is None:
            await self._open()
        return SshClient(host,
                         port,
                         username=username,
                         password=password,
                         tunnel=self._conn,
                         verbose=self.verbose)

    async def cmd(self, cmd, timeout=5.0, delimiter=None, term_type=None):
        """Send a command.  The response is returned."""
        await self._open()
        if self.stdin is None:
            try:
                self.stdin, self.stdout, self.stderr = await self._conn.open_session(
                    term_type=term_type, encoding='latin-1')  # PPA-1457
                if not delimiter:
                    # this sometimes helps "kick" a connection so the first
                    # command works when it wouldn't otherwise.  saw this with
                    # some ignitenet dropbear servers.  only done when
                    # the default delimiter is used because it can break the
                    # regular delimiter.
                    self.stdin.write('\n')
                    await asyncio.sleep(self.init_wait)
                    await self.stdout.read(2048)
                else:
                    r = await asyncio.wait_for(self.stdout.readuntil(delimiter),
                                               timeout=timeout)
            except asyncio.CancelledError:
                raise
            except Exception as err:
                logging.warning('Error establishing session to %s: %s', self.host, err)
                raise

        if self.verbose > 2:
            print('#!>{}'.format(cmd))
        token = delimiter or '--PRESEEM--'
        if delimiter is None:  # default behavior is we add a token to output
            cmd += '; echo -n {}\n'.format(token)
        self.stdin.write(cmd)
        r = await asyncio.wait_for(self.stdout.readuntil(token), timeout=timeout)
        if not delimiter:
            r = r[:-11]
        if self.verbose > 2:
            print('#!<{}'.format(r))
        if self.record is not None:
            rr = r[:-len(delimiter)] if delimiter else r
            self.record.append([cmd, rr])
        return r

    async def scmd(self, cmd, timeout=5.0):
        """Simpler version of cmd: Send a command, the response is returned."""
        async with self.sem:
            await self._open()
            r = await self._conn.run(cmd, timeout=timeout)
            if self.record is not None:
                self.record.append([cmd, r.stdout])
            return r.stdout

    async def get_brmacs(self, br_name):
        """Dict of Intf:[list of Brmacs] pairs is returned."""
        port_map = {}
        intf_brmacs = {}

        if not br_name:
            raise ValueError("br_name must not be empty")
        if not isinstance(br_name, str):
            raise TypeError("br_name must be a string")

        # Get port_map
        showstp = await self.scmd(rf"brctl showstp {br_name} | grep '^\w* (\w)'")
        showstp = showstp.split('\n') if showstp else []

        if not showstp:
            raise NoBridgeError(f'Invalid bridge {br_name}')

        for port in showstp:
            data = port.split() if port else []
            if len(data) == 2:
                vir_id = data[1].strip('()')
                port_map[vir_id] = data[0]
                intf_brmacs[data[0]] = []

        # Get brmacs
        showmacs = await self.scmd(f"brctl showmacs {br_name}")
        showmacs = showmacs.split('\n')[1:] if showmacs else []
        for row in showmacs:
            row = row.split() if row else []
            if row:
                # Sample row:       ['2', '60:22:32:39:4d:32', 'no', '0.83']
                if row[2] == 'no':  # not local
                    intf = port_map.get(row[0])
                    if intf:
                        intf_brmacs.get(intf).append(row[1])

        return intf_brmacs



from datetime import datetime
from functools import partial
from uuid import UUID

from google.protobuf.descriptor import FieldDescriptor
from google.protobuf.timestamp_pb2 import Timestamp


def field_is_repeated(field):
    """Check if a protobuf field is repeated, compatible with old and new protobuf."""
    try:
        return field.is_repeated
    except AttributeError:
        return field.label == FieldDescriptor.LABEL_REPEATED


def _py_result(model, r, del_uuid):
    """Convert the result of a model operation to a python value."""

    def pb2py(pb, cls):
        """convert a protobuf to a python object."""
        fields = {}
        for x in pb.DESCRIPTOR.oneofs_by_name.values():
            if not x.name.startswith('_'):  # used for optional attributes, ignore
                fields[x.name] = pb.WhichOneof(x.name)  # handle "oneof" attributes
        for f in pb.DESCRIPTOR.fields_by_name:
            try:
                v = getattr(pb, f, None) if pb.HasField(f) else None
            except ValueError:  # HasField only works for optional and submessage
                v = getattr(pb, f, None)
            if isinstance(v, Timestamp):
                v = v.ToNanoseconds() / 1000000000
            fields[f] = v
        return cls(**fields)

    if isinstance(r, list):
        if r:
            cls = model._types.get(r[0].DESCRIPTOR.name)
            return [pb2py(x, cls) for x in r]
        return []
    else:
        cls = model._types.get(r.DESCRIPTOR.name)
        if cls:
            return pb2py(r, cls)
        if del_uuid:
            return None
        if len(r.DESCRIPTOR.fields) == 0:
            # This is just an empty response, we simply return None.
            return
        return r


def _pb_request(reqtype, **kwargs):
    """Setup a GRPC request protobuf for the indicated request and arguments."""
    req = reqtype()
    for k, v in kwargs.items():
        f = reqtype.DESCRIPTOR.fields_by_name[k]
        if v is None and not (f.message_type
                              and f.message_type.name.startswith('Nullable')):
            continue  # allow args to be passed a None to avoid setting them.
        if f.message_type:  # this is a nested message, not a simple type
            if isinstance(v, datetime
                          ) and f.message_type.full_name == 'google.protobuf.Timestamp':
                getattr(req, k).FromDatetime(v)  # accept a python datetime object
            elif f.message_type.name.startswith('Nullable'):
                # special case where we have a OneOf for optional values
                if v is None:
                    setattr(getattr(req, k), 'null', 0)  # NullValue.NULL_VALUE
                else:
                    setattr(getattr(req, k), 'data', v)
            else:
                if isinstance(v, list):
                    # A list of a nested type
                    for x in v:
                        lv = getattr(req, k).add()
                        for nk, nt in f.message_type.fields_by_name.items():
                            nv = getattr(x, nk)
                            try:
                                if not x.HasField(nk):
                                    continue  # skip Optional fields that are not set
                            except ValueError:
                                pass  # non-optional field
                            if nt.message_type:
                                getattr(lv, nk).MergeFrom(nv)
                            elif nv is not None:
                                field = x.DESCRIPTOR.fields_by_name.get(nk)
                                if field and field_is_repeated(field):
                                    getattr(lv, nk).MergeFrom(nv)
                                else:
                                    setattr(lv, nk, nv)
                else:
                    # A plain nested type
                    try:
                        getattr(req, k).CopyFrom(v)
                    except AttributeError:
                        getattr(req, k).MergeFrom(v)

        else:
            if isinstance(v, list):
                getattr(req, k).MergeFrom(v)
            else:
                setattr(req, k, v)
    return req


async def op(model, op, reqtype, delete=False, _pythonize=True, handler=None, **kwargs):
    """Wrapper function to perform a GRPC operation on the model. (asyncio)"""
    req = _pb_request(reqtype, **kwargs)
    company_uuid = kwargs.get('company_uuid') or (
        kwargs.get('uuid') if reqtype.__module__ == 'company_pb2' else None)
    md = (('company-uuid', str(UUID(bytes=company_uuid))), ) if company_uuid else None
    r = await model._client._grpc_op(op,
                                     req,
                                     collect_results=True,
                                     metadata=md,
                                     handler=handler)
    del_uuid = kwargs.get('uuid') if delete else None
    if handler:
        return r
    return _py_result(model, r, del_uuid) if _pythonize else r


def op_sync(model, op, reqtype, delete=False, **kwargs):
    """Wrapper function to perform a GRPC operation on the model. (non-asyncio)"""
    company_uuid = kwargs.get('company_uuid') or (
        kwargs.get('uuid') if reqtype.__module__ == 'company_pb2' else None)
    md = (('company-uuid', str(UUID(bytes=company_uuid))), ) if company_uuid else None
    req = _pb_request(reqtype, **kwargs)
    r = op(req, metadata=md)
    try:
        iterator = iter(r)  # this is to detect whether its a sequence
        r = [x for x in iterator]  # this does network i/o
    except TypeError:  # not an iterable result, ok
        pass
    del_uuid = kwargs.get('uuid') if delete else None
    return _py_result(model, r, del_uuid)


def _extract_service_name(stubcls):
    prefixes = ['ServiceStub', 'ApiStub']
    return next((stubcls[:-len(prefix)]
                 for prefix in prefixes if stubcls.endswith(prefix)), stubcls)


def load_api(model, wrapper, pmod, gmod):
    stubs = [x for x in dir(gmod) if x.endswith('Stub')]
    if not stubs:
        return
    stubcls = stubs[0]
    svcstub = getattr(gmod, stubcls)(model._client._channel)
    svcname = _extract_service_name(stubcls)
    # Create a new python class for this model object type
    cls = type(svcname, (BaseModelObject, ), {})
    setattr(model, svcname, cls)
    model._types[svcname] = cls
    # Create python methods for the services
    for _, s in pmod.DESCRIPTOR.services_by_name.items():
        for method in s.methods:
            reqtype = getattr(pmod, method.input_type.name)
            reqns = type(reqtype.DESCRIPTOR.name, (BaseModelObject, ), {})
            setattr(cls, reqtype.DESCRIPTOR.name, reqns)
            op = getattr(svcstub, method.name)
            setattr(
                cls, method.name,
                partial(wrapper,
                        model=model,
                        op=op,
                        delete=method.name == 'Delete',
                        reqtype=reqtype))
            if method.output_type and not method.output_type.name.endswith('Response'):
                if method.output_type.name not in model._types:
                    model._types[method.output_type.name] = type(
                        method.output_type.name, (BaseModelObject, ), {})
            for nt in reqtype.DESCRIPTOR.nested_types:
                # Create a nested object class
                nc = type(nt.name, (BaseModelObject, ),
                          {f: None
                           for f in nt.fields_by_name})
                setattr(reqns, nt.name, nc)


# XXX once we require Python 3.7, this should be a dataclass.
class BaseModelObject:
    """Base class for all pythonized model objects."""

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        attrs = {
            x: str(UUID(bytes=y)) if y and x.endswith('uuid') else y
            for x, y in self.__dict__.items()
        }
        return f'{type(self).__name__}({attrs})'

    def __eq__(self, other):
        if other.__class__ is not self.__class__:
            return NotImplemented
        return self.__dict__ == other.__dict__

    def __hash__(self):
        return hash((x[1] for x in sorted(self.__dict__.items())))

    def replace(self, **kwargs):
        """Return a copy of this object with some fields updated."""
        args = self.__dict__.copy()
        args.update(kwargs)
        return self.__class__(**args)



"""
Unit test for presee-model.
The preseem-module dynamically loads model protobufs, so to get code coverage
there is no need to test every one - plus we don't want to have to update this
test case when the model protobufs are changed but the preseem_model code is
not changed.  Therefore we test one part of the model (Company) for coverage.

This uses the grpc_testing module (pip install grpcio-testing).
I used the following examples to figure out how to use it:
https://github.com/grpc/grpc/blob/master/src/python/grpcio_tests/tests/testing/_client_test.py
"""
import asyncio
from concurrent.futures import ThreadPoolExecutor
import unittest
from uuid import uuid4

import google.protobuf.timestamp_pb2
import grpc
import grpc_testing

from preseem import PreseemGrpcClient, PreseemModel
from preseem_protobuf.model import company_pb2, company_pb2_grpc

class StubGrpcClient(PreseemGrpcClient):
    def __init__(self):
        super().__init__(api_key=None)

    async def connect(self):
        self._channel = grpc_testing.channel(company_pb2.DESCRIPTOR.services_by_name.values(), grpc_testing.strict_real_time())


class TestPreseemModel(unittest.TestCase):
    def setUp(self):
        self.thread_pool = ThreadPoolExecutor(1)
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.client = StubGrpcClient()
        self.model = PreseemModel(self.client, cache=True)
        self.loop.run_until_complete(self.model.init())

    def tearDown(self):
        self.loop.close()
        self.thread_pool.shutdown(wait=True)

    def _delete(self, cfg):
        return self.loop.run_until_complete(self.model.delete(cfg))

    def _list(self):
        return self.loop.run_until_complete(self.model.list())

    def _listen(self, cbk=None):
        return self.loop.run_until_complete(self.model.listen(cbk))

    def _await(self, co):
        return self.loop.run_until_complete(co)

    async def _do_rpc(self, opname, op, response):
        """Execute an RPC using the grpc-testing unit testing framework."""
        task_op = asyncio.create_task(op)
        await asyncio.sleep(0)  # yield the eventloop
        md, req, rpc = self.client._channel.take_unary_stream(company_pb2.DESCRIPTOR.services_by_name['CompanyService'].methods_by_name[opname])
        rpc.send_initial_metadata(())
        if isinstance(response, list):
            for x in response:
                rpc.send_response(x)
        else:
            rpc.send_response(response)
        rpc.terminate((), grpc.StatusCode.OK, '')
        return await asyncio.wait_for(task_op, timeout=1)

    def _test_op(self, opname, op, response):
        return self.loop.run_until_complete(self._do_rpc(opname, op, response))

    def test_test_1(self):
        """Test the List operation."""
        t = 1618245749.950572
        ts = google.protobuf.timestamp_pb2.Timestamp()
        ts.FromNanoseconds(int(t * 1000000000))
        companies = [
            company_pb2.Company(id=1, uuid=uuid4().bytes, name='company1', inactive=False, created_at=ts),
            company_pb2.Company(id=2, uuid=uuid4().bytes, name='company2', inactive=True, created_at=ts)
        ]
        r = self._test_op('List', self.model.Company.List(), companies)
        self.assertIsInstance(r, list)
        self.assertEqual(len(r), 2)
        for i in range(2):
            self.assertEqual(r[i].id, companies[i].id)
            self.assertEqual(r[i].uuid, companies[i].uuid)
            self.assertEqual(r[i].name, companies[i].name)
            self.assertEqual(r[i].inactive, companies[i].inactive)
            self.assertEqual(r[i].created_at, t)



"""
Fake model used for testing and verification.
"""
import asyncio
import itertools
import pytricia
from preseem import NetworkMetadataEntity, NetworkMetadataReference
from . import NetworkMetadataLowercaseIndex


class FakeNetworkMetadataModel(object):

    def __init__(self, cbk=None, refs_by_type=None, indexes=None):
        self._cbk = cbk  # optional callback for updates
        self.refs_by_type = refs_by_type
        self.refs = {}  # public: current references
        self.nms = {}  # public: current entities
        self.refs_last_updated = {}
        self.pt4 = pytricia.PyTricia(32)
        self._index = {x.type: {} for x in indexes or []}
        self._index_cfg = {x.type: {} for x in indexes or []}
        for index in indexes or []:
            self._index[index.type][index.attribute] = {}
            self._index_cfg[index.type][index.attribute] = index

    def close(self):
        pass

    def is_loading(self):
        return False

    async def list_nm(self, resolve_refs=False):
        return list(self.nms.values())

    async def set_nm(self, entity):
        assert isinstance(entity, NetworkMetadataEntity)
        self.nms[entity.network_prefix] = entity
        if ':' not in entity.network_prefix:
            try:
                self.pt4[entity.network_prefix] = entity
            except Exception as err:
                logging.warning('Error setting nm entity %s: %s', entity.network_prefix,
                                err)
                del self.nm_syn[nm.network_prefix]
        if self._cbk:
            await self._cbk(entity, deleted=False)

    async def del_nm(self, entity):
        assert isinstance(entity, NetworkMetadataEntity)
        try:
            del self.nms[entity.network_prefix]
            try:
                del self.pt4[entity.network_prefix]
            except KeyError:
                pass
            if self._cbk:
                await self._cbk(entity, deleted=True)
        except AttributeError:
            pass

    def get_ref_index(self, reftype, attr, val=None):
        return self._index[reftype][attr].get(val)

    async def list_refs(self):
        if self.refs_by_type:
            return list(
                itertools.chain.from_iterable([x.values() for x in self.refs.values()]))
        else:
            return list(self.refs.values())

    async def set_ref(self, ref):
        assert isinstance(ref, NetworkMetadataReference)
        if self.refs_by_type:
            refs = self.refs.get(ref.type)
            if refs is None:
                refs = self.refs[ref.type] = {}
            cur = refs.get(ref.value)
            refs[ref.value] = ref
            index = self._index.get(ref.type)
            if index is not None:
                for ia, ir in index.items():
                    is_lower = isinstance(self._index_cfg[ref.type][ia],
                                          NetworkMetadataLowercaseIndex)
                    if ia == '_LOWER':
                        refs = ir.get(None)
                        if refs is None:
                            refs = ir[None] = {}
                        refs[ref.value.lower()] = ref
                        continue
                    if cur:
                        # Remove the old ref from any indexes
                        attr = cur.attributes.get(ia)
                        if attr:
                            if is_lower:
                                attr = attr.lower()
                            refs = ir.get(attr)
                            if refs:
                                refs.pop(ref.value, None)
                    attr = ref.attributes.get(ia)
                    if attr:
                        if is_lower:
                            attr = attr.lower()
                        refs = ir.get(attr)
                        if refs is None:
                            refs = ir[attr] = {}
                        refs[ref.value] = ref
        else:
            self.refs[(ref.type, ref.value)] = ref
        self.refs_last_updated[ref.type] = asyncio.get_event_loop().time()
        if self._cbk:
            await self._cbk(ref, deleted=False)

    async def del_ref(self, ref):
        assert isinstance(ref, NetworkMetadataReference)
        try:
            if self.refs_by_type:
                ref = self.refs[ref.type].pop(ref.value)
                if not self.refs[ref.type]:
                    del self.refs[ref.type]
                index = self._index.get(ref.type)
                if index is not None:
                    for ia, ir in index.items():
                        is_lower = isinstance(self._index_cfg[ref.type][ia],
                                              NetworkMetadataLowercaseIndex)
                        if ia == '_LOWER':
                            refs = ir.get(None)
                            if refs:
                                refs.pop(ref.value.lower(), None)
                            continue
                        attr = ref.attributes.get(ia)
                        if attr:
                            if is_lower:
                                attr = attr.lower()
                            refs = ir.get(attr)
                            if refs:
                                refs.pop(ref.value, None)
            else:
                del self.refs[(ref.type, ref.value)]
            self.refs_last_updated[ref.type] = asyncio.get_event_loop().time()
            if self._cbk:
                await self._cbk(ref, deleted=True)
        except AttributeError:
            pass

    async def update(self):
        pass
